"use strict"

class Bybit extends Exchange {
	static meta = {
		aliases: [
			"BYBIT",
		],
		endpoint: "https://api.bybit.com",
//		endpoint: "https://api.bytick.com",
		fields: {
			public: {
				label: "API Key",
			},
			private: {
				label: "API Secret",
			},
		},
		name: "Bybit v2",
		patterns: [],
		permissions: {
			origins: [
				"https://api.bybit.com/*"
//				"https://api.bytick.com/*"
			],
			permissions: [
				"webRequest",
				"webRequestBlocking",
			],
		},
		subscriptions: {
			active: [
				"e3fdd77de3a682bfcfd8dc3000a25a5b",
				"4db6261b067dabf8ffb0924c18d95d06",
			],
			inactive: [
				"bybit1",
				"bybit2",
			],
		},
		ignoreCode: [
			34015,
			34036,
			130127,
			30083, // position mode not modified
			30084, // Isolated not modified
			// 10005, // Permission denied, please check your API key permissions
		],
		ignoreMsg: [
			"cannot set leverage which is same to the old leverage",
			"leverage not modified",
			"unknown result",
			"not modified",
			"position mode not modified", // 30083
			"isolated not modified", // 30084
			//"Permission denied, please check your API key permissions", // 10005
		],
		recvWindow: 30 * 1000,
		website: "https://www.bybit.com/register?affiliate_id=777",
		desc: "Bybit is the safest, fastest, most transparent, and user friendly Bitcoin and Ethereum trading platform offering cryptocurrency perpetual contracts.",
		margin: true, spot: true, posMode: true, marginType: true,
		testSymbol: "BTCUSDTS",
		defaultSymbol: "BTC",
		isInverse: sym => sym && sym.endsWith("USD"),
		isLinear: sym => sym && sym.endsWith("USDT"),
		isFuture: sym => sym && !sym.endsWith("USD") && !sym.endsWith("USDT"),
		isSpot: sym => sym && sym.endsWith('S')
	}
	constructor() {
		super(Bybit.meta)
	}

	*req(method, resource, params, sign = true) {
		if (sign) {
			const credentials = this.getCredentials("private")
			params = params || {}
			params.api_key = credentials.public
			params.timestamp = this.getTimestamp()
			params.recv_window = this.meta.recvWindow
			params = serialize(params, null, true, false)

			let sha = new jsSHA("SHA-256", "TEXT")
			sha.setHMACKey(credentials.private, "TEXT")
			sha.update(params)
			params = params.replace(/\+/g, '%2B') + "&sign=" + sha.getHMAC("HEX")
			if (method === 'DELETE') {
				resource += '?' + params
				params = null
			}
		} else this.hasAccess()

		this.removeOrigin(true)
		const response = yield request.bind(this, method, this.meta.endpoint + resource, params, null)

		if (!response) {
			throw new Error("Unknown connection error with API endpoint!")
		}
		if (response.ret_code !== 0 || response.ext_code) {
			if (!this.meta.ignoreCode.includes(Number(response.ret_code)) && !this.meta.ignoreMsg.includesNoCase(response.ret_msg)) {
				throw new Error(`${response.ret_msg || "Unknown connection error with API endpoint!"} (Code ${response.ret_code})`)
			}
		}
		return response.result
	}

	*time() {
		const data = yield request.bind(this, 'GET', this.meta.endpoint + "/v2/public/time", null, null)
		return Math.round(data.time_now * 1000)
	}

	*account(currency, isMargin) {
		let balances = {}, hasSpotPermission = true
		const isTest = this.cmd.id === 'TEST'

		if (!isMargin || isTest) {
			try {
				const spot = (yield* this.req('GET', "/spot/v1/account")).balances
				spot && spot.forEach && spot.forEach(item => {
					const coin = item.coin.toUpperCase()
					if (Number(item.total) || coin == currency) balances[coin] = {
						available: Number(item.free),
						balance: Number(item.total)
					}
				})
			} catch (ex) {
				// Invalid API-key, IP, or permissions for action. (Code -2015)
				if (!isTest) throw(ex)
				hasSpotPermission = false
				this.warn('seems account has no spot trading permissions!')
			}
		}
		if (isMargin || isTest) {
			try {
				const wallets = yield* this.req('GET', "/v2/private/wallet/balance", !isTest && currency ? {coin: currency} : {})
				for (const coin in wallets) {
					const wallet = wallets[coin]
					const equity = wallet.equity || (wallet.wallet_balance + wallet.unrealised_pnl) || 0
					if (equity) {
						balances[isMargin ? coin : coin+marginSuffix] = {
							available: wallet.available_balance || (wallet.wallet_balance - wallet.used_margin) || 0,
							balance: equity,
							wallet: wallet.wallet_balance
						}
					}
				}
			} catch (ex) {
				// Permission denied, please check your API key permissions. (Code 10005)
				if (!isTest || !hasSpotPermission) throw(ex)
				this.warn('seems account has no contract trading permissions!')
			}
		}
		isTest && currency && (balances[currency] = balances[currency] || {available: 0, balance: 0})
		return balances
	}

	*symbolInfo(symbol) {
		const symbolCache = this.getSymbolCache()
		if (symbolCache._refresh) {
			const response = yield* this.req('GET', "/v2/public/symbols", undefined, false)
			response.forEach(info => {
				const tickSize = info.price_filter && info.price_filter.tick_size
				const qtyStep = info.lot_size_filter && info.lot_size_filter.qty_step
				info.pricePrecision = decimals(tickSize) + (!tickSize.toString().endsWith('1') ? tickSize % 1 : 0)
				info.quantityPrecision = decimals(qtyStep) + (!qtyStep.toString().endsWith('1') ? qtyStep % 1 : 0)
				info.minimumTradeSize = info.lot_size_filter && info.lot_size_filter.min_trading_qty
				info.maximumTradeSize = info.lot_size_filter && info.lot_size_filter.max_trading_qty
				info.maximumPostSize = info.lot_size_filter && info.lot_size_filter.post_only_max_trading_qty
				info.maxLeverage = info.leverage_filter && info.leverage_filter.max_leverage
				if (info.quote_currency === "USDT") {
					info.type = "USDT Perpetual"
					info.isLinear = true
				} else if (info.alias != info.name) {
					info.type = "Inverse Futures"
					info.isFuture = true
					symbolCache[info.alias] = info
				} else {
					info.type = "Inverse Perpetual"
					info.isInverse = true
				}
				symbolCache[info.name] = info
			})
			symbolCache.BTCUSDT = symbolCache.BTCUSDT || {
				name: "BTCUSDT",
				quote_currency: "USDT",
				type: "USDT Perpetual",
				isLinear: true,
				pricePrecision: 1.1,
				quantityPrecision: 3,
				minimumTradeSize: 0.001,
			}

			const spot = yield* this.req('GET', "/spot/v1/symbols", undefined, false)
			spot.forEach(info => {
				info.pricePrecision = decimals(info.minPricePrecision)
				info.basePrecision = decimals(info.basePrecision)
				info.quotePrecision = decimals(info.quotePrecision)
				info.baseMinSize = Number(info.minTradeQuantity)
				info.baseMaxSize = Number(info.maxTradeQuantity)
				info.quoteMinSize = Number(info.minTradeAmount)
				info.quoteMaxSize = Number(info.maxTradeAmount)
				info.type = "Spot"
				info.isSpot = true
				symbolCache[info.name+'S'] = info
			})

			this.updatedSymbolCache()
		}
		if (!symbol) return symbolCache

		if (!symbolCache[symbol]) {
			throw new Error(`Unknown market symbol: ${symbol} – Use the 'Symbols' button in the exchange settings or alert editor to browse for valid symbols!`)
		}
		return symbolCache[symbol]
	}

	*symbolTicker(symbol) {
		const ticker = this.meta.isSpot(symbol) ? (yield* this.req('GET', "/spot/quote/v1/ticker/book_ticker", {symbol: symbol.slice(0, -1)})) : 
			(yield* this.req('GET', "/v2/public/tickers", {symbol: symbol}, false))[0]

		return ticker && {
			ask: Number(ticker.askPrice || ticker.ask_price),
			bid: Number(ticker.bidPrice || ticker.bid_price),
			mid: (Number(ticker.askPrice || ticker.ask_price) + Number(ticker.bidPrice || ticker.bid_price)) / 2,
			last: Number(ticker.last_price),
			mark: Number(ticker.mark_price),
			index: Number(ticker.index_price)
		}
	}

	*getPosMode(market) {
		const posMode = {Buy: 10, Sell: 10, hedge: false, cross: false}
		try {
			if (market.isLinear || market.isFuture) {
				let result = yield* this.req('GET', market.isFuture ? "/futures/private/position/list" : "/private/linear/position/list", {symbol: market.name})
				market.isFuture && (result = result.map(pos => pos.data))
				result.forEach(pos => {
					if (pos.position_idx === 0 && pos.side === "None") {
						posMode.Buy = posMode.Sell = pos.leverage
					} else {
						posMode[pos.position_idx === 2 ? "Sell" : pos.position_idx === 1 ? "Buy" : pos.side] = pos.leverage
					}
					if (pos.position_idx !== 0) posMode.hedge = true
					if (!pos.is_isolated) posMode.cross = true
				})
			} else {
				const pos = yield* this.req('GET', "/v2/private/position/list", {symbol: market.name})
				posMode.Buy = posMode.Sell = pos.leverage
				if (!pos.is_isolated) posMode.cross = true
			}
		} catch (ex) {}
		return posMode
	}

	*trade(cmd, ohlc = {}) {
		const market = yield* this.symbolInfo(cmd.s)
		cmd._sym = market.isSpot ? cmd.s : market.name
		const ticker = (cmd.pc && !cmd.up && this.getPrice(cmd._sym)) || (yield* this.symbolTicker(cmd._sym))
		if (!ticker) {
			throw new ReferenceError(`Ticker ${market.name} is not available!`)
		}
		this.checkSlip(cmd, ohlc, cmd._sym, ticker)
		if (cmd.up) {
			this.updatePrice(cmd._sym, ticker)
			if (!cmd.b && !cmd.ub) {
				return ticker
			}
		}

		const currency = cmd.currency = market.isSpot ? (cmd.isBid ? market.quoteCurrency : market.baseCurrency) : (market.isLinear ? market.quote_currency : market.base_currency)
		if (!market.isSpot) cmd.currency += marginSuffix
		let balance = (cmd.bc || cmd.y === "possize") && !cmd.ub && this.getBalance(currency, !market.isSpot)
		const balances = !balance && (yield* this.account(currency, !market.isSpot))
		balance = balance || balances[currency]
		if (!balance) {
			throw new ReferenceError(`Account Balance ${currency} not available! (${market.isSpot ? "Spot" : "Derivatives"} Wallet)`)
		}
		if (cmd.minb && balance.available < cmd.minb._().reference(balance.balance).resolve()) {
			throw new Error("Not enough balance available! (minb was specified)")
		} else if (cmd.maxb && balance.available > cmd.maxb._().reference(balance.balance).resolve()) {
			throw new Error("Too much balance available! (maxb was specified)")
		}
		if (!cmd.bc || cmd.ub) {
			this.updateBalance(balances, currency, !market.isSpot)
			if (cmd.ub) {
				return balance
			}
		}

		const first = ohlc[cmd.pr] || ticker[cmd.pr] || ((cmd.isBid && cmd.t !== "market") || (cmd.isAsk && cmd.t === "market") ? ticker.bid : ticker.ask)
		let price = cmd.fp ? cmd.fp.resolve(market.pricePrecision) : cmd.p._().relative(first).minmax(cmd.minp, cmd.maxp).resolve(market.pricePrecision)
		if (market.price_filter && market.price_filter.min_price && price < market.price_filter.min_price) {
			this.warn(`price below instrument minimum of ${market.price_filter.min_price} – use minp= to set minimum!`)
		} else if (price <= 0) {
			this.warn("your p parameter might be incorrect! (price of 0 calculated, use minp= to set minimum)")
		}

		ohlc[cmd.lr] && this.info(`using custom leverage multiplier: ${ohlc[cmd.lr]} (${cmd.lr})`)
		const available = cmd.y === "equity" ? balance.balance : balance.available
		let cf = 1.0
		const posMode = market.isSpot ? {} : (yield* this.getPosMode(market))
		const doMode = (cmd.pm === "normal" && posMode.hedge) || (cmd.pm === "hedge" && !posMode.hedge)
		const cross = cmd.mt ? cmd.mt === "crossed" : posMode.cross || cmd.l === 0
		let leverage = cmd._lev = market.isSpot ? 1 : cmd.lc || (cmd.l === 0 ? 10 : cmd.l) * (ohlc[cmd.lr] || 1) || posMode[cmd.isBid ? "Buy" : "Sell"]
		if (leverage > market.maxLeverage) {
			cmd.l && (cf = market.maxLeverage / leverage)
			leverage = cmd._lev = market.maxLeverage
			this.warn(`leverage limited to ${leverage}x by instrument!`)
		}
		const contracts = ohlc[cmd.y] || (
				market.isSpot ? available / (cmd.isBid && cmd.t !== "market" ? price : 1) : 
				market.isLinear ? available / price : available * price) * leverage * cf

		ohlc[cmd.qr] && this.info(`using custom quantity multiplier: ${ohlc[cmd.qr]} (${cmd.qr})`)
		let order = {
			side: cmd.isBid ? "Buy" : "Sell",
			symbol: market.name,
			order_type: cmd.t === "market" ? "Market" : "Limit",
			qty: cmd.q._().mul(cmd.u === "currency" && !cmd.q.wasPercent() ? (market.isLinear ? 1 / price : price) * leverage * cf : 1)
				.reference(contracts).mul(ohlc[cmd.qr] || 1).minmax(cmd.minq, cmd.maxq).resolve(market.isSpot ? (cmd.isBid && cmd.t === "market" ? market.quotePrecision : market.basePrecision) : market.quantityPrecision),
			time_in_force: ({fok: "FillOrKill", ioc: "ImmediateOrCancel", post: "PostOnly"})[cmd.t] || "GoodTillCancel",
			close_on_trigger: false,
			reduce_only: false
		}
		if (cmd.lq && ohlc.left) {
			this.info("using leftover quantity from last order: "+ohlc.left)
			order.qty = ohlc.left
		}
		const maxQty = cmd.t === "post" && market.maximumPostSize || market.maximumTradeSize
		if (market.minimumTradeSize && order.qty < market.minimumTradeSize) {
			this.warn(`order quantity below instrument minimum of ${market.minimumTradeSize} – use minq= to set minimum!`)
		} else if (maxQty && order.qty > maxQty) {
			this.warn(`order quantity above instrument maximum of ${maxQty} – use maxq= to set maximum!`)
		} else if (order.qty <= 0) {
			this.warn("your q parameter might be incorrect! (quantity of 0 calculated, use minq= to set minimum)")
		}
		if (market.isLinear || market.isFuture) {
			order.position_idx = (!doMode && !posMode.hedge) || (doMode && posMode.hedge) ? 0 : cmd.isBid ? (!cmd.r ? 1 : 2) : (!cmd.r ? 2 : 1)
		}
		if (cmd.t !== "market") order.price = price

		if (market.isSpot) {
			order.type = cmd.t === "market" ? "MARKET" : "LIMIT"
			delete order.order_type
			order.timeInForce = ({fok: "FOK", ioc: "IOC"})[cmd.t] || "GTC"
			delete order.time_in_force
			delete order.close_on_trigger
			delete order.reduce_only
		} else {
			if (cmd.so) {
				cmd.soref && cmd.soref !== "price" && this.info(`using custom stop reference: ${ohlc[cmd.soref] || ticker[cmd.soref] || "NOT FOUND!"} (${cmd.soref})`)
				cmd.soref !== "price" && (price = (cmd.soref && (ohlc[cmd.soref] || ticker[cmd.soref])) || first)
				order.stop_px = cmd.so._().relative(price).resolve(market.pricePrecision)
				order.base_price = ticker.last
				order.trigger_by = ucfirst(cmd.sr) + "Price"
			}
			if (cmd.tp) {
				cmd.tpref && cmd.tpref !== "price" && this.info(`using custom TP reference: ${ohlc[cmd.tpref] || ticker[cmd.tpref] || "NOT FOUND!"} (${cmd.tpref})`)
				const stop = cmd.tpref === "price" ? price : ((cmd.tpref && (ohlc[cmd.tpref] || ticker[cmd.tpref])) || first)
				order.take_profit = cmd.tp._().relative(stop).resolve(market.pricePrecision)
				order.tp_trigger_by = ucfirst(cmd.sr) + "Price"
			}
			if (cmd.sl) {
				cmd.slref && cmd.slref !== "price" && this.info(`using custom SL reference: ${ohlc[cmd.slref] || ticker[cmd.slref] || "NOT FOUND!"} (${cmd.slref})`)
				const stop = cmd.slref === "price" ? price : ((cmd.slref && (ohlc[cmd.slref] || ticker[cmd.slref])) || first)
				order.stop_loss = cmd.sl._().relative(stop).resolve(market.pricePrecision)
				order.sl_trigger_by = ucfirst(cmd.sr) + "Price"
			}
			//cmd.so && (cmd.tp || cmd.sl) && !market.isLinear && this.warn("TP/SL 'on fill' only supported for USDT markets!")

			if (cmd.r) order.close_on_trigger = order.reduce_only = true
		}
		if (cmd.id) order[market.isSpot ? 'orderLinkId' : 'order_link_id'] = this.uniqueID(cmd.id)

		this.info(`placing ${currency} order${market.isSpot?'':` @ ${leverage}x${cross?' cross':' isolated'}${order.position_idx?' (hedged)':''}`}: `+stringify(order))
		if (cmd.d) {
			this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return order
		}

		if ((market.isLinear || market.isFuture) && doMode) {
			this.info(`setting position mode to ${cmd.pm}...`)
			try {
				yield* this.req('POST', market.isFuture ? "/futures/private/position/switch-mode" : "/private/linear/position/switch-mode", 
					{symbol: market.name, mode: market.isFuture ? (cmd.pm !== "hedge" ? 0 : 3) : (cmd.pm !== "hedge" ? "MergedSingle" : "BothSide")})
			} catch(ex) {
				this.warn(`couldn't change position mode, please change manually! (Error: ${ex.message})`)
			}
		} else if (doMode) {
			this.warn(`this market does not support different position modes!`)
		}

		const posSide = order.position_idx === 2 ? "Sell" : order.position_idx === 1 ? "Buy" : order.side
		const doLev = cmd.l !== undefined && leverage != posMode[posSide]
		const doType = cross != posMode.cross
		if (!market.isSpot && (doLev || doType)) {
			this.info(`setting contract leverage to ${leverage}x${cross?' cross':' isolated'}...`)
			posMode[posSide] = leverage
			if ((!doMode && !posMode.hedge) || (doMode && cmd.pm !== "hedge")) {
				posMode.Buy = posMode.Sell = leverage
			}
			try {
				if (market.isInverse) {
					if (doType) {
						yield* this.req('POST', "/v2/private/position/switch-isolated", 
							{symbol: market.name, is_isolated: !cross, buy_leverage: posMode.Buy, sell_leverage: posMode.Sell})
					} else {
						yield* this.req('POST', "/v2/private/position/leverage/save", {symbol: market.name, leverage: posMode[posSide], leverage_only: true})
					}
				} else {
					yield* this.req('POST', 
						market.isLinear ? (doType ? "/private/linear/position/switch-isolated" : "/private/linear/position/set-leverage") : 
										  (doType ? "/futures/private/position/switch-isolated" : "/futures/private/position/leverage/save"), 
						{symbol: market.name, position_idx: cmd.isBid ? (!cmd.r ? 1 : 2) : (!cmd.r ? 2 : 1), is_isolated: !cross, buy_leverage: posMode.Buy, sell_leverage: posMode.Sell})
				} 
			} catch(ex) {
				this.warn(`couldn't set leverage, please set manually! (Error: ${ex.message})`)
			}
		}

		if (maxQty && order.qty > maxQty) {
			let leftQty = order.qty, result = null, i = 1
			do {
				if (i > 1 && opt.splitDelay) {
					this.info(`waiting for ${opt.splitDelay}sec (split order delay)...`)
					while((yield* sleep(opt.splitDelay)) == null);
				}
				order.qty = Math.min(leftQty, maxQty)
				if (cmd.id) order[market.isSpot ? 'orderLinkId' : 'order_link_id'] = this.uniqueID(cmd.id, i)
				this.info(`posting split order #${i++} of size ${order.qty}...`)
				result = yield* this.req('POST', market.isSpot ? "/spot/v1/order" : 
					market.isLinear ? (order.stop_px ? "/private/linear/stop-order/create"  : "/private/linear/order/create")  : 
					market.isFuture ? (order.stop_px ? "/futures/private/stop-order/create" : "/futures/private/order/create") : 
									   order.stop_px ? "/v2/private/stop-order/create"      : "/v2/private/order/create", order)
				leftQty -= order.qty
			} while(leftQty > 0)
			return result
		}
		return yield* this.req('POST', market.isSpot ? "/spot/v1/order" : 
			market.isLinear ? (order.stop_px ? "/private/linear/stop-order/create"  : "/private/linear/order/create")  : 
			market.isFuture ? (order.stop_px ? "/futures/private/stop-order/create" : "/futures/private/order/create") : 
							   order.stop_px ? "/v2/private/stop-order/create"      : "/v2/private/order/create", order)
	}

	*ordersCancel(cmd) {
		const market = yield* this.symbolInfo(cmd.s)

		if (!cmd.ch && !cmd.d && !cmd.b && !cmd.fp && !cmd.id && cmd.cm.wasPercent(true) && cmd.r === undefined && 
			(!market.isSpot || (!cmd.has('t') && !cmd.tp && !cmd.sl && !cmd.ts && !cmd.so))) {
			if (market.isSpot) {
				this.info(`canceling all ${cmd.s} limit orders in bulk!`)
				yield* this.req('DELETE', "/spot/order/batch-cancel", {symbol: market.name})
			} else {
				// 1 = stop, 2 = limit, 3 = all
				const type = !cmd.tp && !cmd.sl && !cmd.ts && !cmd.so && cmd.t !== "close" ? (!cmd.has('t') ? 3 : 2) : 1
				if (type > 1) {
					this.info(`canceling all ${market.name} limit orders in bulk!`)
					yield* this.req('POST', market.isLinear ? "/private/linear/order/cancel-all" : 
						market.isFuture ? "/futures/private/order/cancelAll" : "/v2/private/order/cancelAll", {symbol: market.name})
				}
				if (type !== 2) {
					this.info(`canceling all ${market.name} stop orders in bulk!`)
					yield* this.req('POST', market.isLinear ? "/private/linear/stop-order/cancel-all" : 
						market.isFuture ? "/futures/private/stop-order/cancelAll" : "/v2/private/stop-order/cancelAll", {symbol: market.name})
				}
			}
			return true
		}

		let orders = []
		if (market.isSpot) {
			orders = (yield* this.req('GET', "/spot/v1/open-orders", {symbol: market.name})) || []
		} else if (market.isLinear) {
			orders = (yield* this.req('GET', "/private/linear/order/search", {symbol: market.name})) || []
			orders = orders.concat(((yield* this.req('GET', "/private/linear/stop-order/search", {symbol: market.name})) || []))
		} else {
			orders = (yield* this.req('GET', market.isFuture ? "/futures/private/order" : "/v2/private/order", {symbol: market.name})) || []
			orders = orders.concat(((yield* this.req('GET', market.isFuture ? "/futures/private/stop-order" : "/v2/private/stop-order", {symbol: market.name})) || [])
				.filter(order => order.stop_order_id = order.order_id))
		}
		if (!orders.length) {
			this.info("no orders open!")
			return false
		}

		const ids = (cmd.id || "").split(',')
		const link = market.isSpot ? 'orderLinkId' : 'order_link_id'
		const hasType = cmd.has('t') && (cmd.t === "limit" || cmd.t === "market") && ucfirst(cmd.t)
		orders = orders.filter(order => {
			if ((cmd.isBid && order.side !== 'Buy') || (cmd.isAsk && order.side !== 'Sell')) {
				return false
			}
			if (cmd.fp && cmd.fp.compare(order.price)) {
				return false
			}
			if (hasType && order.order_type !== hasType) {
				return false
			}
			if ((cmd.t === "open" && order.stop_order_id) || ((cmd.t === "close" || cmd.tp || cmd.sl || cmd.ts || cmd.so) && !order.stop_order_id)) {
				return false
			}
			if (cmd.r !== undefined && !cmd.r != !order.reduce_only) {
				return false
			}
			if (cmd.id && (!order[link] || !order[link].startsWithAny(ids, true))) {
				return false
			}
			return true
		})

		if (cmd.cm._().reference(orders.length).getMax() < orders.length) {
			const sort = {
				"newest": 	[market.isLinear ? "created_time" : "created_at", true],
				"oldest": 	[market.isLinear ? "created_time" : "created_at", false],
				"highest": 	["price", true],
				"lowest": 	["price", false],
				"biggest": 	["qty", true],
				"smallest": ["qty", false]
			}
			cmd.cmo === 'random' ? shuffle(orders) : sortByIndex(orders, sort[cmd.cmo][0], sort[cmd.cmo][1])
			orders = orders.slice(0, cmd.cm.resolve(0))
		}
		if(!orders.length) {
			this.info("open orders don't match any criteria!")
			return []
		}

		this.info((cmd.ch ? "orders found: " : "canceling orders: ")+stringify(orders))
		if (cmd.ch || cmd.d) {
			!cmd.ch && this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return orders
		}

		let cancelOrders = []
		for (const order of orders) {
			const result = market.isSpot ? (yield* this.req('DELETE', "/spot/v1/order", {orderId: order.orderId})) : 
				(yield* this.req('POST', 
					market.isLinear ? (order.stop_order_id ? "/private/linear/stop-order/cancel"  : "/private/linear/order/cancel")  : 
					market.isFuture ? (order.stop_order_id ? "/futures/private/stop-order/cancel" : "/futures/private/order/cancel") : 
								order.stop_order_id ? "/v2/private/stop-order/cancel"      : "/v2/private/order/cancel", {
					symbol: market.name, order_id: order.order_id, stop_order_id: order.stop_order_id
				}))
			result && (result.order_id || result.stop_order_id) && cancelOrders.push(order)
		}
		return cancelOrders
	}

	*positionsClose(cmd, pos, i, ohlc = {}, ticker, market) {
		ticker = ticker || (cmd.pc && !cmd.up && this.getPrice(market.name)) || (yield* this.symbolTicker(market.name))
		if (!ticker) {
			throw new ReferenceError(`Ticker ${market.name} is not available!`)
		}
		ticker.pos = pos.entry_price
		ticker.liq = pos.liq_price
		this.checkSlip(cmd, ohlc, market.name, ticker)
		cmd.up && this.updatePrice(market.name, ticker)

		const first = ohlc[cmd.pr] || ticker[cmd.sl || cmd.tp || cmd.ts || cmd.so ? 'pos' : cmd.pr] || ((cmd.isBid && cmd.t !== "market") || (cmd.isAsk && cmd.t === "market") ? ticker.bid : ticker.ask)
		let price = cmd.p._().relative(first).minmax(cmd.minp, cmd.maxp).resolve(market.pricePrecision)
		if (cmd.fp) {
			price = cmd.fp.resolve(market.pricePrecision)
		}
		if (market.price_filter && market.price_filter.min_price && price < market.price_filter.min_price) {
			this.warn(`price below instrument minimum of ${market.price_filter.min_price} – use minp= to set minimum!`)
		} else if (price <= 0) {
			this.warn("your p parameter might be incorrect! (price of 0 calculated, use minp= to set minimum)")
		}

		const contracts = pos.quantity
		ohlc[cmd.qr] && this.info(`using custom quantity multiplier: ${ohlc[cmd.qr]} (${cmd.qr})`)
		let order = {
			side: pos.side === 'Buy' ? 'Sell' : 'Buy',
			symbol: market.name,
			order_type: cmd.t === "market" ? "Market" : "Limit",
			qty: cmd.q._().mul(cmd.u === "currency" && !cmd.q.wasPercent() ? (market.isLinear ? 1 / price : price) * (pos.leverage || 1) : 1)
				.reference(contracts).mul(ohlc[cmd.qr] || 1).minmax(cmd.minq, cmd.maxq).resolve(market.quantityPrecision),
			time_in_force: ({fok: "FillOrKill", ioc: "ImmediateOrCancel", post: "PostOnly"})[cmd.t] || "GoodTillCancel",
			close_on_trigger: cmd.r === undefined ? true : cmd.r,
			reduce_only: cmd.r === undefined ? true : cmd.r
		}
		if (cmd.t !== "market") order.price = price

		if (((cmd.tp || cmd.sl) && (!cmd.q.wasPercent(true) || cmd.t === "limit" || cmd.id)) || (cmd.so && !cmd.ts)) {
			const ref = (cmd.tp && cmd.tpref) || (cmd.sl && cmd.slref) || (cmd.so && cmd.soref)
			ref && ref !== "price" && this.info(`using custom stop reference: ${ohlc[ref] || ticker[ref] || "NOT FOUND!"} (${ref})`)
			const stop = ref === "price" ? price : ((ref && (ohlc[ref] || ticker[ref])) || first)
			order.stop_px = (cmd.tp || cmd.sl || cmd.so)._().relative(stop).resolve(market.pricePrecision)
			order.base_price = ticker.last
			order.trigger_by = ucfirst(cmd.sr) + "Price"
		} else if ((cmd.tp || cmd.sl || cmd.ts) && !cmd.ch) {
			order = {symbol: market.name}
			if (market.isLinear) order.side = pos.side
			if (market.isLinear || market.isFuture) order.position_idx = pos.position_idx
			if (cmd.tp) {
				cmd.tpref && cmd.tpref !== "price" && this.info(`using custom TP reference: ${ohlc[cmd.tpref] || ticker[cmd.tpref] || "NOT FOUND!"} (${cmd.tpref})`)
				const stop = cmd.tpref === "price" ? price : ((cmd.tpref && (ohlc[cmd.tpref] || ticker[cmd.tpref])) || first)
				order.take_profit = cmd.tp._().relative(stop).resolve(market.pricePrecision)
				order.tp_trigger_by = ucfirst(cmd.sr) + "Price"
			}
			if (cmd.sl) {
				cmd.slref && cmd.slref !== "price" && this.info(`using custom SL reference: ${ohlc[cmd.slref] || ticker[cmd.slref] || "NOT FOUND!"} (${cmd.slref})`)
				const stop = cmd.slref === "price" ? price : ((cmd.slref && (ohlc[cmd.slref] || ticker[cmd.slref])) || first)
				order.stop_loss = cmd.sl._().relative(stop).resolve(market.pricePrecision)
				order.sl_trigger_by = ucfirst(cmd.sr) + "Price"
			}
			if (cmd.ts) {
				cmd.soref && cmd.soref !== "price" && this.info(`using custom trigger reference: ${ohlc[cmd.soref] || ticker[cmd.soref] || "NOT FOUND!"} (${cmd.soref})`)
				const act = cmd.soref === "price" ? price : ((cmd.soref && (ohlc[cmd.soref] || ticker[cmd.soref])) || first)
				if (cmd.so) if (market.isFuture || market.isInverse) {
					order.new_trailing_active = cmd.so._().relative(act).resolve(market.pricePrecision)
				} else {
					this.warn("does not support a trigger price for trailing stops on this market!")
				}
				cmd.tsref && cmd.tsref !== "price" && this.info(`using custom TS reference: ${ohlc[cmd.tsref] || ticker[cmd.tsref] || "NOT FOUND!"} (${cmd.tsref})`)
				const stop = cmd.tsref === "price" ? price : ((cmd.tsref && (ohlc[cmd.tsref] || ticker[cmd.tsref])) || act || first)
				order.trailing_stop = cmd.ts._().reference(stop).abs().resolve(market.pricePrecision)
			}
			cmd.id && this.warn("does not support custom IDs for Trading Stop position attributes – use so= instead to place a normal stop order!")
			this.info("setting Trading Stop: "+stringify(order))
			if (cmd.d) {
				this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
				return order
			}
			return (yield* this.req('POST', 
				market.isLinear ? "/private/linear/position/trading-stop" : 
				market.isFuture ? "/futures/private/position/trading-stop" : 
								  "/v2/private/position/trading-stop", order)) || order
		}
		if (market.isLinear || market.isFuture) order.position_idx = pos.position_idx
		if (cmd.id) order.order_link_id = this.uniqueID(cmd.id, i)

		!cmd.ch && this.info(`placing Close Position order @ ${pos.leverage}x: `+stringify(order))
		if (cmd.ch || cmd.d) {
			!cmd.ch && this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return order
		}

		const maxQty = (cmd.t === "post" && market.maximumPostSize) || market.maximumTradeSize
		if (maxQty && order.qty > maxQty) {
			let leftQty = order.qty, result = null, i = 1
			do {
				if (i > 1 && opt.splitDelay) {
					this.info(`waiting for ${opt.splitDelay}sec (split order delay)...`)
					while((yield* sleep(opt.splitDelay)) == null);
				}
				order.qty = Math.min(leftQty, maxQty)
				if (cmd.id) order['order_link_id'] = this.uniqueID(cmd.id, i)
				this.info(`posting split order #${i++} of size ${order.qty}...`)
				result = yield* this.req('POST', 
					market.isLinear ? (order.stop_px ? "/private/linear/stop-order/create"  : "/private/linear/order/create")  : 
					market.isFuture ? (order.stop_px ? "/futures/private/stop-order/create" : "/futures/private/order/create") : 
									   order.stop_px ? "/v2/private/stop-order/create"      : "/v2/private/order/create", order)
				leftQty -= order.qty
			} while(leftQty > 0)
			return result
		}
		return yield* this.req('POST', 
			market.isLinear ? (order.stop_px ? "/private/linear/stop-order/create"  : "/private/linear/order/create")  : 
			market.isFuture ? (order.stop_px ? "/futures/private/stop-order/create" : "/futures/private/order/create") : 
							   order.stop_px ? "/v2/private/stop-order/create"      : "/v2/private/order/create", order)
	}

	*positionsCloseAll(cmd, ohlc, positions) {
		if (!positions) {
			const sym = cmd.s !== '*' && cmd.s
			const market = sym && (yield* this.symbolInfo(sym))
			let inverse = !market || market.isInverse ? (yield* this.req('GET', "/v2/private/position/list", market && {symbol: market.name})) : []
			inverse && inverse.symbol && (inverse = [inverse])
			if (!inverse || !inverse.filter) {
				this.info("unable to get position info!")
				return false
			}
			const linear = !market || market.isLinear ? (yield* this.req('GET', "/private/linear/position/list", market && {symbol: market.name})) : []
			if (!linear || !linear.filter) {
				this.info("unable to get position info!")
				return false
			}
			const future = !market || market.isFuture ? (yield* this.req('GET', "/futures/private/position/list", market && {symbol: market.name})) : []
			if (!future || !future.filter) {
				this.info("unable to get position info!")
				return false
			}
			positions = inverse.concat(linear, future.map(pos => pos.data)).map(e => e.data || e).filter(pos => pos.side !== 'None' && pos.size)
			if (!positions.length) {
				this.info("no positions open!")
				return false
			}
		}

		const tickers = {}
		for (const pos of positions) {
			if (!tickers[pos.symbol]) {
				tickers[pos.symbol] = ((cmd.up || ((cmd.minpl || cmd.maxpl || !cmd.s) && (pos.unrealised_pnl === undefined || this.meta.isFuture(pos.symbol)))) 
					&& (yield* this.symbolTicker(pos.symbol))) || (cmd.pc && this.getPrice(pos.symbol))
			}
		}

		const symbols = yield* this.symbolInfo()
		positions = positions.filter(pos => {
			if ((cmd.isBid && pos.side !== 'Buy') || (cmd.isAsk && pos.side !== 'Sell')) {
				return false
			}
			if (cmd.fp && cmd.fp.compare(pos.entry_price)) {
				return false
			}
			if (cmd.l && pos.leverage !== cmd.l) {
				return false
			}
			const market = symbols[pos.symbol] || {}
			cmd._sym = market.name
			pos.currency = cmd.currency = (market.isLinear ? market.quote_currency : market.base_currency) + marginSuffix
			pos.quantity = Math.abs(pos.size)
			const ticker = tickers[pos.symbol]
			if ((pos.unrealised_pnl === undefined || market.isFuture) && ticker) {
				pos.unrealised_pnl = (pos.side === 'Buy' ? ticker.mark - pos.entry_price : pos.entry_price - ticker.mark) * 
					(market.isFuture ? pos.position_value / pos.entry_price : pos.quantity)
			}
			if (ticker) pos.mark_price = pos.mark_price || ticker.mark
			pos.unrealised_pnl = toPrecision(pos.unrealised_pnl || 0, 8)
			pos.pnl = pos.unrealised_pnl / pos.position_margin / pos.leverage
			pos._sizeCoin = market.isLinear ? pos.quantity / pos.leverage : pos.position_margin
			pos._sizeUSD = market.isLinear ? pos.position_margin : pos.quantity / pos.leverage
			pos.avgEntryPrice = toPrecision(pos.entry_price || 0, market.pricePrecision)
			if (!this.checkPL(cmd, ohlc, pos, balance => (market.isLinear ? balance / pos.entry_price : balance * pos.entry_price) * pos.leverage, market.quantityPrecision)) {
				return false
			}
			return true
		})

		if (cmd.cm._().reference(positions.length).getMax() < positions.length) {
			const sort = {
				"newest": 	["created_at", true],
				"oldest": 	["created_at", false],
				"highest": 	["entry_price", true],
				"lowest": 	["entry_price", false],
				"biggest": 	["quantity", true],
				"smallest": ["quantity", false]
			}
			cmd.cmo === 'random' ? shuffle(positions) : sortByIndex(positions, sort[cmd.cmo][0], sort[cmd.cmo][1])
			positions = positions.slice(0, cmd.cm.resolve(0))
		}
		if(!positions.length) {
			this.info("open positions don't match any criteria!")
			return []
		} else if (cmd.ch) {
			this.info(`${positions.length} position${positions.length > 1 ? 's' : ''} found!`)
			return positions
		}

		let closeOrders = [], i = 0
		for (const pos of positions) {
			const order = yield* this.positionsClose(cmd, pos, i++, ohlc, tickers[pos.symbol], symbols[pos.symbol])
			if (order) {
				closeOrders.push(order)
				Object.assign(closeOrders.last(), Object.without(pos, ["quantity", "size", "side", "take_profit", "stop_loss", "trailing_stop"]))
			}
		}
		return closeOrders
	}
}

Broker.add(new Bybit())
